import os

from pathlib import Path
from typing import List, Optional, OrderedDict, Union

from supervisely import (
    Api,
    ProjectMeta,
    VolumeAnnotation,
    batched,
    generate_free_name,
    is_development,
    logger,
)
from supervisely.convert.base_converter import BaseConverter
from supervisely.io.fs import get_file_ext, get_file_name_with_ext
from supervisely.volume.volume import ALLOWED_VOLUME_EXTENSIONS, read_nrrd_serie_volume


class VolumeConverter(BaseConverter):
    allowed_exts = ALLOWED_VOLUME_EXTENSIONS

    class Item(BaseConverter.BaseItem):
        def __init__(
            self,
            item_path: str,
            ann_data: str = None,
            shape: tuple = None,
            custom_data: Optional[dict] = None,
            volume_meta: Union[dict, OrderedDict] = None,
            mask_dir: str = None,
            interpolation_dir: str = None,
        ):
            self._path: str = item_path
            self._ann_data: str = ann_data
            if volume_meta is None:
                sitk_volume, meta = read_nrrd_serie_volume(item_path)
                self._volume_meta = meta
            else:
                self._volume_meta: Union[dict, OrderedDict] = volume_meta
            if shape is None:
                self._shape: tuple = (
                    self._volume_meta["dimensionsIJK"]["y"],
                    self._volume_meta["dimensionsIJK"]["x"],
                    self._volume_meta["dimensionsIJK"]["z"],
                )
            else:
                self._shape: tuple = shape

            self._mask_dir: str = mask_dir
            self._interpolation_dir: str = interpolation_dir
            self._type: str = "volume"
            self._custom_data: dict = custom_data if custom_data is not None else {}

        @property
        def volume_meta(self) -> Union[dict, OrderedDict]:
            return self._volume_meta

        @property
        def mask_dir(self) -> str:
            return self._mask_dir

        @property
        def interpolation_dir(self) -> str:
            return self._interpolation_dir

        @volume_meta.setter
        def volume_meta(self, meta: Union[dict, OrderedDict]) -> None:
            self._volume_meta = meta

        @mask_dir.setter
        def mask_dir(self, mask_dir: str) -> None:
            self._mask_dir = mask_dir

        @interpolation_dir.setter
        def interpolation_dir(self, interpolation_dir: str) -> None:
            self._interpolation_dir = interpolation_dir

        def create_empty_annotation(self) -> VolumeAnnotation:
            return VolumeAnnotation(self._volume_meta)

    def __init__(self, input_data: str, labeling_interface: str):
        self._input_data: str = input_data
        self._meta: ProjectMeta = None
        self._items: List[self.Item] = []
        self._labeling_interface: str = labeling_interface
        self._converter = self._detect_format()
        self._batch_size: int = 1

    @property
    def format(self):
        return self._converter.format

    @property
    def ann_ext(self):
        return None

    @property
    def key_file_ext(self):
        return None

    def get_meta(self) -> ProjectMeta:
        return self._meta

    def get_items(self) -> List[BaseConverter.BaseItem]:
        return self._items

    @staticmethod
    def validate_ann_file(ann_path, meta=None):
        return False

    def upload_dataset(
        self,
        api: Api,
        dataset_id: int,
        batch_size: int = 1,
        log_progress=True,
    ):
        """Upload converted data to Supervisely"""

        meta, renamed_classes, renamed_tags = self.merge_metas_with_conflicts(api, dataset_id)

        existing_names = set([vol.name for vol in api.volume.get_list(dataset_id)])

        if log_progress:
            progress, progress_cb = self.get_progress(self.items_count, "Uploading volumes...")
        else:
            progress_cb = None

        for batch in batched(self._items, batch_size=batch_size):
            item_names = []
            item_paths = []
            anns = []
            mask_dirs = []
            interpolation_dirs = []
            for item in batch:
                ext = get_file_ext(item.path)
                if ext.lower() != ext:
                    new_volume_path = Path(item.path).with_suffix(ext.lower()).as_posix()
                    os.rename(item.path, new_volume_path)
                    item.path = new_volume_path
                item.name = get_file_name_with_ext(item.path)
                item.name = generate_free_name(
                    existing_names, item.name, with_ext=True, extend_used_names=True
                )
                item_names.append(item.name)
                item_paths.append(item.path)
                anns.append(item.ann_data)
                mask_dirs.append(item.mask_dir)
                interpolation_dirs.append(item.interpolation_dir)

                ann = self.to_supervisely(item, meta, renamed_classes, renamed_tags)
                anns.append(ann)

            vol_infos = api.volume.upload_nrrd_series_paths(
                dataset_id,
                item_names,
                item_paths,
            )
            vol_ids = [vol_info.id for vol_info in vol_infos]
            if all(ann is not None for ann in anns):
                api.volume.annotation.upload_paths(
                    vol_ids, anns, meta, interpolation_dirs=interpolation_dirs, mask_dirs=mask_dirs
                )

            if log_progress:
                progress_cb(len(batch))

        if log_progress:
            if is_development():
                progress.close()
        logger.info(f"Dataset ID:{dataset_id} has been successfully uploaded.")



import os
from typing import List

from supervisely import Annotation, ProjectMeta, logger
from supervisely.convert.base_converter import AvailableImageConverters
from supervisely.convert.image.image_converter import ImageConverter
from supervisely.convert.image.pascal_voc import pascal_voc_helper
from supervisely.io.fs import (
    dir_exists,
    dirs_filter,
    file_exists,
    get_file_name,
    list_files_recursively,
    remove_junk_from_dir,
)


class PascalVOCConverter(ImageConverter):
    class Item(ImageConverter.Item):
        def __init__(
            self,
            *args,
            **kwargs,
        ):
            super().__init__(*args, **kwargs)
            self._segm_path = None
            self._inst_path = None

        @property
        def segm_path(self) -> str:
            return self._segm_path

        @property
        def inst_path(self) -> str:
            return self._inst_path

        @segm_path.setter
        def segm_path(self, segm_path: str) -> None:
            self._segm_path = segm_path

        @inst_path.setter
        def inst_path(self, inst_path: str) -> None:
            self._inst_path = inst_path

    def __init__(self, input_data: str, labeling_interface: str) -> None:
        self._input_data: str = input_data
        self._items: List[ImageConverter.Item] = []
        self._meta: ProjectMeta = None
        self.color2class_name = None
        self.with_instances = False
        self._imgs_dir = None
        self._segm_dir = None
        self._inst_dir = None
        self._labeling_interface = labeling_interface

    def __str__(self) -> str:
        return AvailableImageConverters.PASCAL_VOC

    @property
    def ann_ext(self) -> str:
        return ".xml"

    def validate_ann_file(self, ann_path: str, meta: ProjectMeta) -> bool:
        pass

    def validate_format(self) -> bool:
        detected_ann_cnt = 0

        def check_function(dir_path):
            possible_image_dir_names = ["JPEGImages", "Images", "images", "imgs", "img"]
            possible_segm_dir_names = [
                "SegmentationClass",
                "segmentation",
                "segmentations",
                "Segmentation",
                "Segmentations",
                "masks",
                "segm",
            ]
            if not any([dir_exists(os.path.join(dir_path, p)) for p in possible_image_dir_names]):
                return False
            if not any([dir_exists(os.path.join(dir_path, p)) for p in possible_segm_dir_names]):
                return False
            for d in possible_image_dir_names:
                if dir_exists(os.path.join(dir_path, d)):
                    self._imgs_dir = os.path.join(dir_path, d)
                    break
            for d in possible_segm_dir_names:
                if dir_exists(os.path.join(dir_path, d)):
                    self._segm_dir = os.path.join(dir_path, d)
                    break
            return True

        possible_pascal_voc_dir = [d for d in dirs_filter(self._input_data, check_function)]
        if len(possible_pascal_voc_dir) == 0:
            return False
        if len(possible_pascal_voc_dir) > 1:
            logger.warn("Multiple Pascal VOC directories not supported")
            return False

        possible_pascal_voc_dir = possible_pascal_voc_dir[0]
        remove_junk_from_dir(possible_pascal_voc_dir)

        colors_file = os.path.join(possible_pascal_voc_dir, "colors.txt")
        obj_classes, color2class_name = pascal_voc_helper.read_colors(colors_file)
        self.color2class_name = color2class_name
        for p in ["SegmentationObject", "instances", "objects"]:
            self.with_instances = dir_exists(os.path.join(possible_pascal_voc_dir, p))
            if self.with_instances:
                self._inst_dir = os.path.join(possible_pascal_voc_dir, p)
                break
        self._meta = ProjectMeta(obj_classes=obj_classes)

        images_list = list_files_recursively(self._imgs_dir, valid_extensions=self.allowed_exts)

        # create Items
        self._items = []
        for image_path in images_list:
            item = self.Item(image_path)
            item_name_noext = get_file_name(item.name)
            segm_path = os.path.join(self._segm_dir, item_name_noext + ".png")
            if file_exists(segm_path):
                item.segm_path = segm_path
                detected_ann_cnt += 1
            if self.with_instances:
                inst_path = os.path.join(self._inst_dir, item_name_noext + ".png")
                if file_exists(inst_path):
                    item.inst_path = inst_path
            self._items.append(item)
        return detected_ann_cnt > 0

    def to_supervisely(
        self,
        item: ImageConverter.Item,
        meta: ProjectMeta = None,
        renamed_classes: dict = None,
        renamed_tags: dict = None,
    ) -> Annotation:
        if meta is None:
            meta = self._meta

        try:
            item.set_shape()
            ann = pascal_voc_helper.get_ann(item, self.color2class_name, renamed_classes)
            return ann

        except Exception as e:
            logger.warn(f"Failed to convert annotation: {repr(e)}")
            return item.create_empty_annotation()



from typing import List, Literal, NamedTuple
from supervisely.api.api import Api
from supervisely.api.file_api import FileInfo


class CheckpointInfo(NamedTuple):
    app_name: str
    session_id: int
    session_path: str
    session_link: str
    task_type: str
    training_project_name: str
    checkpoints: List[FileInfo]


import supervisely.nn.checkpoints.detectron2 as detectron2
import supervisely.nn.checkpoints.hrda as hrda
import supervisely.nn.checkpoints.mmclassification as mmclassification
import supervisely.nn.checkpoints.mmdetection as mmdetection
import supervisely.nn.checkpoints.mmdetection3 as mmdetection3
import supervisely.nn.checkpoints.mmsegmentation as mmsegmentation
import supervisely.nn.checkpoints.ritm as ritm
import supervisely.nn.checkpoints.unet as unet
import supervisely.nn.checkpoints.yolov5 as yolov5
import supervisely.nn.checkpoints.yolov5_v2 as yolov5_v2
import supervisely.nn.checkpoints.yolov8 as yolov8


def get_list(
    api: Api,
    team_id: int,
    framework: Literal[
        "yolov5",
        "yolov5_v2",
        "yolov8",
        "detectron2",
        "mmdetection",
        "mmdetection3",
        "mmsegmentation",
        "mmclassification",
        "ritm",
        "unet",
        "hrda",
    ],
) -> List[CheckpointInfo]:
    if framework == "yolov5":
        checkpoints = yolov5.get_list(api, team_id)
    elif framework == "yolov5_v2":
        checkpoints = yolov5_v2.get_list(api, team_id)
    elif framework == "yolov8":
        checkpoints = yolov8.get_list(api, team_id)
    elif framework == "detectron2":
        checkpoints = detectron2.get_list(api, team_id)
    elif framework == "mmdetection":
        checkpoints = mmdetection.get_list(api, team_id)
    elif framework == "mmdetection3":
        checkpoints = mmdetection3.get_list(api, team_id)
    elif framework == "mmsegmentation":
        checkpoints = mmsegmentation.get_list(api, team_id)
    elif framework == "mmclassification":
        checkpoints = mmclassification.get_list(api, team_id)
    elif framework == "ritm":
        checkpoints = ritm.get_list(api, team_id)
    elif framework == "unet":
        checkpoints = unet.get_list(api, team_id)
    elif framework == "hrda":
        checkpoints = hrda.get_list(api, team_id)
    else:
        raise NotImplementedError(f"Unknown framework: {framework}")

    return checkpoints



import os
from typing import List

import supervisely.convert.image.masks.image_with_masks_helper as helper
from supervisely import Annotation, ProjectMeta, logger, ObjClass, Bitmap, ObjClassCollection
from supervisely.convert.base_converter import AvailableImageConverters
from supervisely.convert.image.image_converter import ImageConverter
from supervisely.io.fs import file_exists, dirs_with_marker, dir_exists, get_file_name, list_files, dirs_filter, remove_junk_from_dir, get_file_ext
from supervisely.io.json import load_json_file


class ImagesWithMasksConverter(ImageConverter):
    def __init__(self, input_data: str, labeling_interface: str) -> None:
        self._input_data: str = input_data
        self._items: List[ImageConverter.Item] = []
        self._meta: ProjectMeta = None
        self._classes_mapping = {}
        self._labeling_interface = labeling_interface

    def __str__(self):
        return AvailableImageConverters.MASKS

    @property
    def key_file_ext(self) -> str:
        return ".json"

    def validate_key_file(self, key_file_path: str) -> bool:
        if not file_exists(key_file_path):
            return False
        try:
            classes_mapping = load_json_file(key_file_path)
            if not isinstance(classes_mapping, dict):
                return False
            cls_list = [ObjClass(cls_name, Bitmap) for cls_name in classes_mapping]
            obj_class_collection = ObjClassCollection(cls_list)
            self._classes_mapping = classes_mapping
            return ProjectMeta(obj_classes=obj_class_collection)
        except Exception as e:
            logger.warn(f"Failed to read obj_class_to_machine_color.json: {repr(e)}")
            return False

    def validate_format(self) -> bool:
        detected_ann_cnt = 0
        possible_dirs = [d for d in dirs_with_marker(self._input_data, helper.COLOR_MAP_FILE_NAME)]
        if len(possible_dirs) == 0:
            return False
        if len(possible_dirs) > 1:
            return False
        project_dir = possible_dirs[0]
        project_dir_name = os.path.basename(project_dir.rstrip("/"))
        colors_file = os.path.join(project_dir, helper.COLOR_MAP_FILE_NAME)
        key_file_result = self.validate_key_file(colors_file)
        if key_file_result is False:
            return False
        self._meta = key_file_result

        # possible_dss 
        def _search_for_dss(dir_path):
            if any([d in os.listdir(dir_path) for d in helper.MASK_DIRS]):
                return True
            return False

        dataset_paths = [d for d in dirs_filter(project_dir, _search_for_dss)]
        dataset_names = [os.path.basename(d.rstrip("/")) for d in dataset_paths]

        self._items = []
        for dataset_path, dataset_name in zip(dataset_paths, dataset_names):
            img_dir = os.path.join(dataset_path, helper.IMAGE_DIR_NAME)
            if not dir_exists(img_dir):
                continue
            names = [get_file_name(file_name) for file_name in os.listdir(img_dir)]
            img_paths = [os.path.join(img_dir, file_name) for file_name in os.listdir(img_dir)]

            machine_masks_dir = os.path.join(dataset_path, helper.MASKS_MACHINE_DIR_NAME)
            instance_masks_dir = os.path.join(dataset_path, helper.MASKS_INSTANCE_DIR_NAME)
            machine_masks_dir_exists = dir_exists(machine_masks_dir)
            instance_masks_dir_exists = dir_exists(instance_masks_dir)
            ann_dir = os.path.join(dataset_path, helper.ANNOTATION_DIR_NAME)
            if dir_exists(ann_dir):
                remove_junk_from_dir(ann_dir)
            if not machine_masks_dir_exists:
                if all([get_file_ext(d) == helper.MASK_EXT for d in os.listdir(ann_dir)]):
                    machine_masks_dir = ann_dir
                    machine_masks_dir_exists = True
            if not instance_masks_dir_exists:
                if all([dir_exists(os.path.join(ann_dir, d)) for d in os.listdir(ann_dir)]):
                    instance_masks_dir = ann_dir
                    instance_masks_dir_exists = True

            for img_name, img_path in zip(names, img_paths):
                item = self.Item(img_path)
                ann_data = {}
                machine_mask = os.path.join(machine_masks_dir, f"{img_name}{helper.MASK_EXT}")
                ann_detected = False
                if machine_masks_dir_exists and file_exists(machine_mask):
                    ann_data["machine_mask"] = machine_mask
                    ann_detected = True
                if instance_masks_dir_exists:
                    mask_dir = os.path.join(instance_masks_dir, img_name)
                    if dir_exists(mask_dir):
                        inst_masks = list_files(mask_dir, valid_extensions=[helper.MASK_EXT])
                        if len(inst_masks) > 0:
                            ann_detected = True
                            ann_data["inst_masks"] = inst_masks
                if ann_detected:
                    detected_ann_cnt += 1
                    item.ann_data = ann_data
                    self._items.append(item)

        return detected_ann_cnt > 0

    def to_supervisely(
        self,
        item: ImageConverter.Item,
        meta: ProjectMeta = None,
        renamed_classes: dict = None,
        renamed_tags: dict = None,
    ) -> Annotation:
        """Convert to Supervisely format."""
        if meta is None:
            meta = self._meta

        ann = item.create_empty_annotation()
        if item.ann_data is None:
            return ann

        try:
            semantic_labels = []
            instance_labels = []
            semantic_mask_path = item.ann_data.get("machine_mask")
            instance_masks_paths = item.ann_data.get("inst_masks")
            if semantic_mask_path is not None:
                semantic_labels = helper.read_semantic_labels(
                    semantic_mask_path, self._classes_mapping, meta.obj_classes, renamed_classes
                )
            if instance_masks_paths is not None:
                instance_labels = helper.read_instance_labels(
                    instance_masks_paths, meta.obj_classes, renamed_classes
                )

            ann = ann.add_labels(labels=semantic_labels + instance_labels)

            return ann
        except Exception as e:
            logger.warn(f"Failed to convert annotation: {repr(e)}")
            return ann



from typing import List

from supervisely import (
    AnyGeometry,
    Bitmap,
    ObjClass,
    Point,
    Polygon,
    Polyline,
    ProjectMeta,
    Rectangle,
    TagMeta,
    TagValueType,
    logger,
)
from supervisely.io.json import load_json_file

SLY_ANN_KEYS = ["size", "framesCount", "frames", "objects", "tags"]


def get_meta_from_annotation(ann_path: str, meta: ProjectMeta) -> ProjectMeta:
    ann_json = load_json_file(ann_path)
    if "annotation" in ann_json:
        ann_json = ann_json["annotation"]
    if not all(key in ann_json for key in SLY_ANN_KEYS):
        logger.warn(f"VideoAnnotation file {ann_path} is not in Supervisely format")
        return meta

    object_key_to_name = {}
    for object in ann_json["objects"]:
        meta = create_tags_from_annotation(meta, object["tags"])
        object_key_to_name[object["key"]] = object["classTitle"]
    for frame in ann_json["frames"]:
        meta = create_classes_from_annotation(frame, meta, object_key_to_name)
    meta = create_tags_from_annotation(meta, ann_json["tags"])
    return meta


def create_tags_from_annotation(meta: ProjectMeta, tags: List[dict]) -> ProjectMeta:
    for tag in tags:
        tag_name = tag["name"]
        tag_value = tag.get("value")
        if tag_value is None:
            tag_meta = TagMeta(tag_name, TagValueType.NONE)
        elif isinstance(tag_value, int) or isinstance(tag_value, float):
            tag_meta = TagMeta(tag_name, TagValueType.ANY_NUMBER)
        else:
            tag_meta = TagMeta(tag_name, TagValueType.ANY_STRING)

        # check existing tag_meta in meta
        existing_tag = meta.get_tag_meta(tag_name)
        if existing_tag is None:
            meta = meta.add_tag_meta(tag_meta)
    return meta


def create_classes_from_annotation(
    frame: dict, meta: ProjectMeta, object_key_to_name: dict
) -> ProjectMeta:
    for fig in frame["figures"]:
        obj_key = fig.get("objectKey")
        if obj_key is None:
            continue
        class_name = object_key_to_name[obj_key]
        geometry_type = fig["geometryType"]
        if geometry_type == Bitmap.geometry_name():
            obj_class = ObjClass(name=class_name, geometry_type=Bitmap)
        elif geometry_type == Rectangle.geometry_name():
            obj_class = ObjClass(name=class_name, geometry_type=Rectangle)
        elif geometry_type == Point.geometry_name():
            obj_class = ObjClass(name=class_name, geometry_type=Point)
        elif geometry_type == Polygon.geometry_name():
            obj_class = ObjClass(name=class_name, geometry_type=Polygon)
        elif geometry_type == Polyline.geometry_name():
            obj_class = ObjClass(name=class_name, geometry_type=Polyline)

        # @TODO: add better check for geometry type, add
        # elif geometry_type == GraphNodes.geometry_name():
        #     geometry_config = None
        #     obj_class = ObjClass(name=class_name, geometry_type=GraphNodes)
        existing_class = meta.get_obj_class(class_name)
        if existing_class is None:
            meta = meta.add_obj_class(obj_class)
        else:
            if existing_class.geometry_type != obj_class.geometry_type:
                obj_class = ObjClass(name=class_name, geometry_type=AnyGeometry)
                meta = meta.delete_obj_class(class_name)
                meta = meta.add_obj_class(obj_class)
    return meta

def rename_in_json(ann_json, renamed_classes=None, renamed_tags=None):
    if renamed_classes:
        for obj in ann_json["objects"]:
            obj["classTitle"] = renamed_classes.get(obj["classTitle"], obj["classTitle"])
            for tag in obj["tags"]:
                tag["name"] = renamed_tags.get(tag["name"], tag["name"])
    if renamed_tags:
        for tag in ann_json["tags"]:
            tag["name"] = renamed_tags.get(tag["name"], tag["name"])
    return ann_json


from supervisely import GraphNodes, Polygon, Rectangle, logger
from supervisely.geometry.graph import KeypointsTemplate, Node
from supervisely.imaging.color import generate_rgb

coco_classes = [
    "person",
    "bicycle",
    "car",
    "motorcycle",
    "airplane",
    "bus",
    "train",
    "truck",
    "boat",
    "traffic light",
    "fire hydrant",
    "stop sign",
    "parking meter",
    "bench",
    "bird",
    "cat",
    "dog",
    "horse",
    "sheep",
    "cow",
    "elephant",
    "bear",
    "zebra",
    "giraffe",
    "backpack",
    "umbrella",
    "handbag",
    "tie",
    "suitcase",
    "frisbee",
    "skis",
    "snowboard",
    "sports ball",
    "kite",
    "baseball bat",
    "baseball glove",
    "skateboard",
    "surfboard",
    "tennis racket",
    "bottle",
    "wine glass",
    "cup",
    "fork",
    "knife",
    "spoon",
    "bowl",
    "banana",
    "apple",
    "sandwich",
    "orange",
    "broccoli",
    "carrot",
    "hot dog",
    "pizza",
    "donut",
    "cake",
    "chair",
    "couch",
    "potted plant",
    "bed",
    "dining table",
    "toilet",
    "tv",
    "laptop",
    "mouse",
    "remote",
    "keyboard",
    "cell phone",
    "microwave",
    "oven",
    "toaster",
    "sink",
    "refrigerator",
    "book",
    "clock",
    "vase",
    "scissors",
    "teddy bear",
    "hair drier",
    "toothbrush",
]


def generate_colors(count):
    colors = []
    for _ in range(count):
        new_color = generate_rgb(colors)
        colors.append(new_color)
    return colors


def get_coordinates(line):
    class_index = int(line[0])
    coords = list(map(float, line[1:]))
    return class_index, coords


def convert_rectangle(img_height, img_width, x_center, y_center, ann_width, ann_height):
    x_center = float(x_center)
    y_center = float(y_center)
    ann_width = float(ann_width)
    ann_height = float(ann_height)

    px_x_center = x_center * img_width
    px_y_center = y_center * img_height

    px_ann_width = ann_width * img_width
    px_ann_height = ann_height * img_height

    left = int(px_x_center - (px_ann_width / 2))
    right = int(px_x_center + (px_ann_width / 2))

    top = int(px_y_center - (px_ann_height / 2))
    bottom = int(px_y_center + (px_ann_height / 2))

    # check if the coordinates are within the image
    left, top = max(0, left), max(0, top)
    right, bottom = min(img_width, right), min(img_height, bottom)

    return Rectangle(top, left, bottom, right)


def convert_polygon(img_height, img_width, *coords):
    exterior = []
    for i in range(0, len(coords), 2):
        x = coords[i]
        y = coords[i + 1]
        px_x = min(img_width, max(0, int(x * img_width)))
        px_y = min(img_height, max(0, int(y * img_height)))
        exterior.append([px_y, px_x])
    return Polygon(exterior=exterior)


def convert_keypoints(img_height, img_width, num_keypoints, num_dims, *coords):
    nodes = []
    step = 3 if num_dims == 3 else 2
    shift = 4
    for i in range(shift, num_keypoints + shift, step):
        x = coords[i]
        y = coords[i + 1]
        visibility = int(coords[i + 2]) if num_dims == 3 else 2
        if visibility in [0, 1]:
            continue  # skip invisible keypoints
        px_x = min(img_width, max(0, int(x * img_width)))
        px_y = min(img_height, max(0, int(y * img_height)))
        node = Node(row=px_x, col=px_y)  # , disabled=v)
        nodes.append(node)
    if len(nodes) > 0:
        return GraphNodes(nodes)


def create_geometry_config(num_keypoints=None):
    i, j = 0, 0
    template = KeypointsTemplate()
    for p in list(range(num_keypoints)):
        template.add_point(label=str(p), row=i, col=j)
        j += 1
        i += 1

    return template



import mimetypes
from pathlib import Path

import magic
from PIL import Image

from supervisely import logger
from supervisely.imaging.image import read, write
from supervisely.io.fs import (
    get_file_ext,
    get_file_name,
    get_file_name_with_ext,
    silent_remove,
)

EXT_TO_CONVERT = [".heic", ".avif", ".heif"]


def validate_image(path: str) -> tuple:
    """Validate image by mimetypes, ext, and remove alpha channel."""

    try:
        ext = get_file_ext(path)
        if ext.lower() == ".nrrd":
            return path
        if ext.lower() in EXT_TO_CONVERT:
            path = convert_to_jpg(path)
            return path
        if ext.lower() != ".mpo":
            name = get_file_name_with_ext(path)
            new_name = validate_mimetypes(name, path)
            if new_name != name:
                img = read(path, remove_alpha_channel=False)
                silent_remove(path)
                path = Path(path).with_name(new_name).as_posix()
                write(path, img, remove_alpha_channel=False)
        return path
    except Exception as e:
        logger.warning(f"Skip image: {repr(e)}", extra={"file_path": path})


def validate_mimetypes(name: str, path: str) -> list:
    """Validate mimetypes for images."""

    mimetypes.add_type("image/webp", ".webp")  # to extend types_map
    mimetypes.add_type("image/jpeg", ".jfif")  # to extend types_map

    mime = magic.Magic(mime=True)
    mimetype = mime.from_file(path)
    file_ext = get_file_ext(path).lower()
    if file_ext in mimetypes.guess_all_extensions(mimetype):
        return name

    new_img_ext = mimetypes.guess_extension(mimetype)
    new_img_name = f"{get_file_name(name)}{new_img_ext}"
    logger.warn(
        f"Image {name} extension doesn't have correct mimetype {mimetype}. "
        f"Image will be converted to {new_img_ext}"
    )

    return new_img_name

def convert_to_jpg(path) -> tuple:
    """Convert image to jpg."""

    # * do not remove folllowing imports, it is used to register avif/heic formats
    import pillow_avif  # pylint: disable=import-error
    from pillow_heif import register_heif_opener  # pylint: disable=import-error

    register_heif_opener()

    new_path = Path(path).with_suffix(".jpeg").as_posix()
    with Image.open(path) as image:
        image.convert("RGB").save(new_path)
    silent_remove(path)
    return new_path



from os.path import join
from typing import List

from supervisely.api.api import Api
from supervisely.io.fs import silent_remove
from supervisely.io.json import load_json_file
from supervisely.nn.checkpoints.checkpoint import CheckpointInfo


def get_list(api: Api, team_id: int) -> List[CheckpointInfo]:
    checkpoints = []
    weights_dir_name = "checkpoints"
    weights_subdir_name = "data"
    info_dir_name = "info"
    training_app_directory = "/mmdetection/"
    if not api.file.dir_exists(team_id, training_app_directory):
        return []
    task_files_infos = api.file.list(team_id, training_app_directory, recursive=False)
    for task_file_info in task_files_infos:
        task_id = task_file_info["name"].split("_")[0]
        project_name = task_file_info["name"].split("_")[1]
        session_link = f"{api.server_address}/apps/sessions/{task_id}"
        path_to_info = join(task_file_info["path"], info_dir_name, "ui_state.json")
        if api.file.exists(team_id, path_to_info):
            api.file.download(team_id, path_to_info, "model_config.json")
            model_config = load_json_file("model_config.json")
            task_type = model_config["task"].replace("_", " ")
            silent_remove("model_config.json")
        else:
            continue
        paths_to_checkpoints = join(task_file_info["path"], weights_dir_name, weights_subdir_name)
        checkpoints_infos = [
            file
            for file in api.file.list(team_id, paths_to_checkpoints, recursive=False)
            if file["name"].endswith(".pth")
        ]
        if len(checkpoints_infos) == 0:
            continue
        checkpoint_info = CheckpointInfo(
            app_name="Train MMDetection",
            session_id=task_id,
            session_path=task_file_info["path"],
            session_link=session_link,
            task_type=task_type,
            training_project_name=project_name,
            checkpoints=checkpoints_infos,
        )
        checkpoints.append(checkpoint_info)
    return checkpoints



import os
from typing import List, Tuple, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from tqdm import tqdm

from supervisely import (
    Annotation,
    Api,
    Progress,
    ProjectMeta,
    TagValueType,
    is_production,
    logger,
)
from supervisely.io.fs import JUNK_FILES, get_file_ext, get_file_name_with_ext


class AvailableImageConverters:
    SLY = "supervisely"
    COCO = "coco"
    YOLO = "yolo"
    PASCAL_VOC = "pascal_voc"
    CSV = "csv"
    MULTISPECTRAL = "multispectral"
    MASKS = "images_with_masks"
    MEDICAL2D = "medical_2d"
    MULTI_VIEW = "multi_view"
    PDF = "pdf"
    CITYSCAPES = "cityscapes"


class AvailableVideoConverters:
    SLY = "supervisely"
    MOT = "mot"
    DAVIS = "davis"


class AvailablePointcloudConverters:
    SLY = "supervisely"
    LAS = "las/laz"
    PLY = "ply"


class AvailablePointcloudEpisodesConverters:
    SLY = "supervisely"


class AvailableVolumeConverters:
    SLY = "supervisely"
    DICOM = "dicom"


class BaseConverter:
    class BaseItem:
        def __init__(
            self,
            item_path: str,
            ann_data: Union[str, dict] = None,
            shape: Union[Tuple, List] = None,
            custom_data: dict = {},
        ):
            self._path: str = item_path
            self._name: str = None
            self._ann_data: Union[str, dict] = ann_data
            self._shape: Union[Tuple, List] = shape
            self._custom_data: dict = custom_data

        @property
        def name(self) -> str:
            return get_file_name_with_ext(self._path)

        @name.setter
        def name(self, name: str) -> None:
            self._name = name

        @property
        def path(self) -> str:
            return self._path

        @path.setter
        def path(self, path: str) -> None:
            self._path = path

        @property
        def ann_data(self) -> Union[str, dict]:
            return self._ann_data

        @ann_data.setter
        def ann_data(self, ann_data: Union[str, dict]) -> None:
            self._ann_data = ann_data

        @property
        def shape(self) -> Union[Tuple, List]:
            return self._shape

        @property
        def custom_data(self) -> dict:
            return self._custom_data

        @custom_data.setter
        def custom_data(self, custom_data: dict) -> None:
            self._custom_data = custom_data

        def set_shape(self, shape) -> None:
            self._shape = shape

        def update_custom_data(self, custom_data: dict) -> None:
            self.custom_data.update(custom_data)

        def update(
            self,
            item_path: str = None,
            ann_data: Union[str, dict] = None,
            shape: Union[Tuple, List] = None,
            custom_data: dict = {},
        ) -> None:
            if item_path is not None:
                self.path = item_path
            if ann_data is not None:
                self.ann_data = ann_data
            if shape is not None:
                self.set_shape(shape)
            if custom_data:
                self.update_custom_data(custom_data)

        def create_empty_annotation(self) -> Annotation:
            raise NotImplementedError()

    def __init__(
        self,
        input_data: str,
        labeling_interface: Literal[
            "default",
            "multi_view",
            "multispectral",
            "images_with_16_color",
            "medical_imaging_single",
        ] = "default",
    ):
        self._input_data: str = input_data
        self._items: List[self.BaseItem] = []
        self._meta: ProjectMeta = None
        self._labeling_interface: str = labeling_interface

    @property
    def format(self) -> str:
        return self.__str__()

    @property
    def items_count(self) -> int:
        return len(self._items)

    @property
    def ann_ext(self) -> str:
        raise NotImplementedError()

    @property
    def key_file_ext(self) -> str:
        raise NotImplementedError()

    def validate_labeling_interface(self) -> bool:
        return self._labeling_interface == "default"

    def validate_ann_file(self, ann_path) -> bool:
        raise NotImplementedError()

    def validate_key_file(self) -> bool:
        raise NotImplementedError()

    def validate_format(self) -> bool:
        """
        Validate format of the input data meets the requirements of the converter. Should be implemented in the subclass.
        Additionally, this method must do the following steps:
            1. creates project meta (if key file file exists) and save it to self._meta
            2. creates items, count detected annotations and save them to self._items
            3. validates annotation files (and genereate meta if key file is missing)

        :return: True if format is valid, False otherwise.
        """
        raise NotImplementedError()

    def get_meta(self) -> ProjectMeta:
        return self._meta

    def get_items(self) -> List[BaseItem]:
        return self._items

    def to_supervisely(self, item: BaseItem, meta: ProjectMeta, *args) -> Annotation:
        """Convert to Supervisely format."""
        return item.create_empty_annotation()

    def upload_dataset(self, api: Api, dataset_id: int, batch_size: int = 50) -> None:
        """Upload converted data to Supervisely"""
        raise NotImplementedError()

    def _detect_format(self):
        found_formats = []
        all_converters = self.__class__.__subclasses__()

        progress, progress_cb = self.get_progress(1, "Detecting annotation format")
        for converter in all_converters:
            if converter.__name__ == "BaseConverter":
                continue
            converter = converter(self._input_data, self._labeling_interface)
            if not converter.validate_labeling_interface():
                continue
            if converter.validate_format():
                logger.info(f"Detected format: {str(converter)}")
                found_formats.append(converter)
                if len(found_formats) > 1:
                    raise RuntimeError(
                        f"Multiple formats detected: {[str(f) for f in found_formats]}. "
                        "Mixed formats are not supported yet."
                    )

        progress_cb(1)

        if len(found_formats) == 0:
            logger.info("Not found any valid annotation formats. Only items will be processed")
            for root, _, files in os.walk(self._input_data):
                for file in files:
                    full_path = os.path.join(root, file)
                    ext = get_file_ext(full_path)
                    if file in JUNK_FILES:
                        continue
                    if ext.lower() in self.allowed_exts:  # pylint: disable=no-member
                        self._items.append(self.Item(full_path))  # pylint: disable=no-member
            if self.items_count == 0:
                raise RuntimeError("No valid items found in the input data")
            return self

        if len(found_formats) == 1:
            return found_formats[0]

    def merge_metas_with_conflicts(
        self, api: Api, dataset_id: int
    ) -> Tuple[ProjectMeta, dict, dict]:

        # get meta1 from project and meta2 from converter
        dataset = api.dataset.get_info_by_id(dataset_id)
        if dataset is None:
            raise RuntimeError(
                f"Dataset ID:{dataset_id} not found. "
                "Please check if the dataset exists and try again."
            )
        meta1_json = api.project.get_meta(dataset.project_id)
        meta1 = ProjectMeta.from_json(meta1_json)
        meta2 = self._meta

        if meta2 is None:
            return meta1, {}, {}

        # merge classes and tags from meta1 (unchanged) and meta2 (renamed if conflict)
        renamed_classes = {}
        renamed_tags = {}
        for new_cls in meta2.obj_classes:
            i = 1
            new_name = new_cls.name
            matched = False
            while meta1.obj_classes.get(new_name) is not None:
                if meta1.obj_classes.get(new_name).geometry_type == new_cls.geometry_type:
                    matched = True
                    break
                new_name = f"{new_cls.name}_{i}"
                i += 1
            if new_name != new_cls.name:
                logger.warn(f"Class {new_cls.name} renamed to {new_name}")
                renamed_classes[new_cls.name] = new_name
            if not matched:
                new_cls = new_cls.clone(name=new_name)
                meta1 = meta1.add_obj_class(new_cls)

        for new_tag in meta2.tag_metas:
            i = 1
            new_name = new_tag.name
            matched = False
            while meta1.tag_metas.get(new_name) is not None:
                if meta1.tag_metas.get(new_name).value_type == new_tag.value_type:
                    if new_tag.value_type != TagValueType.ONEOF_STRING:
                        matched = True
                        break
                    if meta1.tag_metas.get(new_name).possible_values == new_tag.possible_values:
                        matched = True
                        break
                new_name = f"{new_tag.name}_{i}"
                i += 1
            if new_name != new_tag.name:
                logger.warn(f"Tag {new_tag.name} renamed to {new_name}")
                renamed_tags[new_tag.name] = new_name
            if not matched:
                new_tag = new_tag.clone(name=new_name)
                meta1 = meta1.add_tag_meta(new_tag)

        # update project meta
        api.project.update_meta(dataset.project_id, meta1)

        return meta1, renamed_classes, renamed_tags

    def get_progress(
        self,
        total: int,
        message: str = "Processing items...",
        is_size: bool = False,
    ) -> tuple:
        if is_production():
            progress = Progress(message, total, is_size=is_size)
            progress_cb = progress.iters_done_report
        else:
            progress = tqdm(
                total=total, desc=message, unit="B" if is_size else "it", unit_scale=is_size
            )
            progress_cb = progress.update
        return progress, progress_cb



from os.path import basename, join
from typing import List

from supervisely.api.api import Api
from supervisely.io.fs import silent_remove
from supervisely.io.json import load_json_file
from supervisely.nn.checkpoints.checkpoint import CheckpointInfo


def get_list(api: Api, team_id: int) -> List[CheckpointInfo]:
    checkpoints = []
    weights_dir_name = "checkpoints"
    training_app_directory = "/RITM_training/"
    task_type = "instance segmentation"
    if not api.file.dir_exists(team_id, training_app_directory):
        return []
    task_files_infos = api.file.list(team_id, training_app_directory, recursive=False)
    for task_file_info in task_files_infos:
        task_id = task_file_info["name"].split("_")[0]
        project_name = task_file_info["name"].split("_")[1]
        session_link = f"{api.server_address}/apps/sessions/{task_id}"
        paths_to_checkpoints = join(task_file_info["path"], weights_dir_name)
        checkpoints_infos = [
            file
            for file in api.file.list(team_id, paths_to_checkpoints, recursive=False)
            if file["name"].endswith(".pth")
        ]
        if len(checkpoints_infos) == 0:
            continue
        checkpoint_info = CheckpointInfo(
            app_name="Train RITM",
            session_id=task_id,
            session_path=task_file_info["path"],
            session_link=session_link,
            task_type=task_type,
            training_project_name=project_name,
            checkpoints=checkpoints_infos,
        )
        checkpoints.append(checkpoint_info)
    return checkpoints



# isort: skip_file
from supervisely.convert.base_converter import (
    AvailableImageConverters,
    AvailablePointcloudConverters,
    AvailableVideoConverters,
    AvailableVolumeConverters,
    BaseConverter,
)
from supervisely.convert.converter import ImportManager
from supervisely.convert.image.cityscapes.cityscapes_converter import (
    CityscapesConverter,
)

# Image
from supervisely.convert.image.coco.coco_converter import COCOConverter
from supervisely.convert.image.csv.csv_converter import CSVConverter
from supervisely.convert.image.multispectral.multispectral_converter import (
    MultiSpectralImageConverter,
)
from supervisely.convert.image.medical2d.medical2d_converter import Medical2DImageConverter
from supervisely.convert.image.masks.images_with_masks_converter import ImagesWithMasksConverter

from supervisely.convert.image.pascal_voc.pascal_voc_converter import PascalVOCConverter
from supervisely.convert.image.pdf.pdf_converter import PDFConverter
from supervisely.convert.image.sly.sly_image_converter import SLYImageConverter
from supervisely.convert.image.yolo.yolo_converter import YOLOConverter


from supervisely.convert.image.multi_view.multi_view import MultiViewImageConverter

# Pointcloud
from supervisely.convert.pointcloud.sly.sly_pointcloud_converter import SLYPointcloudConverter


# Pointcloud Episodes
from supervisely.convert.pointcloud_episodes.sly.sly_pointcloud_episodes_converter import (
    SLYPointcloudEpisodesConverter,
)
from supervisely.convert.pointcloud.las.las_converter import LasConverter
from supervisely.convert.pointcloud.ply.ply_converter import PlyConverter

# Video
from supervisely.convert.video.mot.mot_converter import MOTConverter
from supervisely.convert.video.sly.sly_video_converter import SLYVideoConverter
from supervisely.convert.volume.dicom.dicom_converter import DICOMConverter

# Volume
from supervisely.convert.volume.sly.sly_volume_converter import SLYVolumeConverter

# Volume
from supervisely.convert.volume.sly.sly_volume_converter import SLYVolumeConverter



from typing import Dict, List, Any, Union, Optional
from supervisely.geometry.graph import GraphNodes
from supervisely.geometry.rectangle import Rectangle
from supervisely.nn.prediction_dto import PredictionKeypoints
from supervisely.annotation.label import Label
from supervisely.nn.inference.inference import Inference
from supervisely.geometry.graph import Node, KeypointsTemplate
from supervisely.project.project_meta import ProjectMeta
from supervisely.annotation.obj_class import ObjClass
from supervisely.imaging import image as sly_image
from supervisely.decorators.inference import _scale_ann_to_original_size, _process_image_path
from supervisely.io.fs import silent_remove
from supervisely.sly_logger import logger
import functools
import numpy as np

try:
    from typing import Literal
except ImportError:
    # for compatibility with python 3.7
    from typing_extensions import Literal


class PoseEstimation(Inference):
    def __init__(
        self,
        model_dir: Optional[str] = None,
        custom_inference_settings: Optional[
            Union[Dict[str, Any], str]
        ] = None,  # dict with settings or path to .yml file
        keypoints_template: Optional[KeypointsTemplate] = None,
        use_gui: Optional[bool] = False,
    ):
        super().__init__(
            model_dir=model_dir,
            custom_inference_settings=custom_inference_settings,
            sliding_window_mode="none",
            use_gui=use_gui,
        )
        self.keypoints_template = keypoints_template

    def get_info(self):
        info = super().get_info()
        info["task type"] = "pose estimation"
        # recommended parameters:
        # info["model_name"] = ""
        # info["checkpoint_name"] = ""
        # info["pretrained_on_dataset"] = ""
        # info["device"] = ""
        return info

    def _get_obj_class_shape(self):
        return GraphNodes

    @property
    def model_meta(self) -> ProjectMeta:
        if self._model_meta is None:
            classes = []
            for name in self.get_classes():
                classes.append(
                    ObjClass(
                        name,
                        self._get_obj_class_shape(),
                        geometry_config=self.keypoints_template,
                    )
                )
            self._model_meta = ProjectMeta(classes)
            self._get_confidence_tag_meta()
        return self._model_meta

    def _create_label(self, dto: PredictionKeypoints):
        obj_class = self.model_meta.get_obj_class(dto.class_name)
        if obj_class is None:
            raise KeyError(
                f"Class {dto.class_name} not found in model classes {self.get_classes()}"
            )
        nodes = []
        for label, coordinate in zip(dto.labels, dto.coordinates):
            x, y = coordinate
            nodes.append(Node(label=label, row=y, col=x))
        label = Label(GraphNodes(nodes), obj_class)
        return label

    def predict(self, image_path: str, settings: Dict[str, Any]) -> List[PredictionKeypoints]:
        raise NotImplementedError("Have to be implemented in child class")

    def predict_raw(self, image_path: str, settings: Dict[str, Any]) -> List[PredictionKeypoints]:
        raise NotImplementedError(
            "Have to be implemented in child class If sliding_window_mode is 'advanced'."
        )

    # pylint: disable=no-self-argument
    def process_image_crop(func):
        """
        Decorator for processing annotation labels before and after inference.
        Crops input image before inference if kwargs['state']['rectangle_crop'] provided
        and then scales annotation back to original image size.
        Keyword arguments:
        :param image_np: Image in numpy.ndarray format (use image_path or image_np, not both)
        :type image_np: numpy.ndarray
        :param image_path: Path to image (use image_path or image_np, not both)
        :type image_path: str
        :raises: :class:`ValueError`, if image_np or image_path invalid or not provided
        :return: Annotation in json format
        :rtype: :class:`dict`
        """

        @functools.wraps(func)
        def wrapper_inference(*args, **kwargs):
            # pylint: disable=not-callable
            settings = kwargs["settings"]

            if "detected_bboxes" in settings:
                ann = func(*args, **kwargs)
                return ann

            rectangle_json = settings.get("rectangle")
            if rectangle_json is None:
                raise ValueError(
                    """Pose estimation task requires target object
                to be outlined in a rectangle before applying neural network."""
                )

            rectangle = Rectangle.from_json(rectangle_json)
            if "image_np" in kwargs.keys():
                image_np = kwargs["image_np"]
                if not isinstance(image_np, np.ndarray):
                    raise ValueError("Invalid input. Image path must be numpy.ndarray")
                original_image_size = image_np.shape[:2]
                image_crop_np = sly_image.crop(image_np, rectangle)
                kwargs["image_np"] = image_crop_np
                ann = func(*args, **kwargs)
                ann = _scale_ann_to_original_size(ann, original_image_size, rectangle)
            elif "image_path" in kwargs.keys():
                image_path = kwargs["image_path"]
                if not isinstance(image_path, str):
                    raise ValueError("Invalid input. Image path must be str")
                image_crop_path, original_image_size = _process_image_path(image_path, rectangle)
                kwargs["image_path"] = image_crop_path
                ann = func(*args, **kwargs)
                ann = _scale_ann_to_original_size(ann, original_image_size, rectangle)
                silent_remove(image_crop_path)
            else:
                raise ValueError("image_np or image_path not provided!")

            return ann

        return wrapper_inference

    @process_image_crop
    def _inference_image_path(
        self,
        image_path: str,
        settings: Dict,
        data_to_return: Dict,  # for decorators
    ):
        logger.debug("Input path", extra={"path": image_path})
        predictions = self.predict(image_path=image_path, settings=settings)
        ann = self._predictions_to_annotation(image_path, predictions)
        return ann



from os.path import basename, join
from typing import List

from supervisely.api.api import Api
from supervisely.io.fs import silent_remove
from supervisely.io.json import load_json_file
from supervisely.nn.checkpoints.checkpoint import CheckpointInfo


# not enough info to implement
def get_list(api: Api, team_id: int) -> List[CheckpointInfo]:
    raise NotImplementedError



# coding: utf-8

# docs
from typing import Dict

from supervisely._utils import batched
from supervisely.api.module_api import ApiField
from supervisely.api.pointcloud.pointcloud_api import PointcloudApi
from supervisely.api.pointcloud.pointcloud_episode_annotation_api import (
    PointcloudEpisodeAnnotationAPI,
)
from supervisely.api.pointcloud.pointcloud_episode_object_api import (
    PointcloudEpisodeObjectApi,
)


class PointcloudEpisodeApi(PointcloudApi):
    """

    API for working with :class:`PointcloudEpisodes<supervisely.pointcloud_episodes.pointcloud_episodes>`.
    :class:`PointcloudEpisodeApi<PointcloudEpisodeApi>` object is immutable.
    Inherits from :class:`PointcloudApi<supervisely.api.pointcloud.PointcloudApi>`.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervise.ly", token="4r47N...xaTatb")

        pcd_epsodes_id = 19373295
        pcd_epsodes_info = api.pointcloud_episode.get_info_by_id(pcd_epsodes_id) # api usage example
    """

    def __init__(self, api):
        super().__init__(api)
        self.annotation = PointcloudEpisodeAnnotationAPI(api)
        self.object = PointcloudEpisodeObjectApi(api)
        self.tag = None

    def _convert_json_info(self, info: dict, skip_missing=True):
        res = super()._convert_json_info(info, skip_missing=skip_missing)
        if res.meta is not None:
            return res._replace(frame=res.meta[ApiField.FRAME])
        else:
            raise RuntimeError(
                "Error with point cloud meta or API version. Please, contact support"
            )

    def get_frame_name_map(self, dataset_id: int) -> Dict:
        """
        Get a dictionary with frame_id and name of pointcloud by dataset id.

        :param dataset_id: :class:`Dataset<supervisely.project.project.Dataset>` ID in Supervisely.
        :type dataset_id: int
        :return: Dict with frame_id and name of pointcloud.
        :rtype: Dict

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dataset_id = 62664
            frame_to_name_map = api.pointcloud_episode.get_frame_name_map(dataset_id)
            print(frame_to_name_map)

            # Output:
            # {0: '001', 1: '002'}
        """

        pointclouds = self.get_list(dataset_id)

        frame_index_to_pcl_name = {}
        if len(pointclouds) > 0 and pointclouds[0].frame is None:
            pointclouds_names = sorted([x.name for x in pointclouds])
            for frame_index, pcl_name in enumerate(pointclouds_names):
                frame_index_to_pcl_name[frame_index] = pcl_name

        else:
            frame_index_to_pcl_name = {x.frame: x.name for x in pointclouds}

        return frame_index_to_pcl_name

    def notify_progress(
        self,
        track_id: int,
        dataset_id: int,
        pcd_ids: list,
        current: int,
        total: int,
    ):
        """
        Send message to the Annotation Tool and return info if tracking was stopped

        :param track_id: int
        :param dataset_id: int
        :param pcd_ids: list
        :param current: int
        :param total: int
        :return: str
        """

        response = self._api.post(
            "point-clouds.episodes.notify-annotation-tool",
            {
                "type": "point-cloud-episodes:fetch-figures-in-range",
                "data": {
                    ApiField.TRACK_ID: track_id,
                    ApiField.DATASET_ID: dataset_id,
                    ApiField.POINTCLOUD_IDS: pcd_ids,
                    ApiField.PROGRESS: {ApiField.CURRENT: current, ApiField.TOTAL: total},
                },
            },
        )
        return response.json()[ApiField.STOPPED]

    def get_max_frame_idx(self, dataset_id: int) -> int:
        """
        Get max frame index for episode by dataset id.
        This method is useful for uploading pointclouds to the episode in parts.

        :param dataset_id: :class:`Dataset<supervisely.project.project.Dataset>` ID in Supervisely.
        :type dataset_id: int
        :return: Max frame index.
        :rtype: int

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dataset_id = 62664
            max_frame = api.pointcloud_episode.get_max_frame(dataset_id)
            print(max_frame)

            # Output:
            # 1
        """

        pointclouds = self.get_list(dataset_id)
        frames = [x.frame for x in pointclouds]
        if len(frames) == 0:
            return None
        max_frame = max(frames)
        return max_frame

    def _upload_bulk_add(
        self,
        func_item_to_kv,
        dataset_id,
        names,
        items,
        metas=None,
        progress_cb=None,
    ):
        if metas is None:
            max_frame = self.get_max_frame_idx(dataset_id)
            if max_frame is None:
                max_frame = range(len(items))
            else:
                max_frame = range(max_frame + 1, max_frame + 1 + len(items))
            metas = [{ApiField.FRAME: i} for i in max_frame]
        else:
            if len(metas) != len(items):
                raise RuntimeError(
                    'Can not match "metas" and "items" lists, len(metas) != len(items)'
                )

            missing_frame_indices = [
                idx for idx, meta in enumerate(metas) if ApiField.FRAME not in meta
            ]
            if len(missing_frame_indices) == len(metas):
                raise RuntimeError("No 'frame' key found in all 'metas'.")
            elif len(missing_frame_indices) > 0:
                missing_frame_names = [names[idx] for idx in missing_frame_indices]
                raise RuntimeError(
                    f"No 'frame' key found in 'metas' for names {missing_frame_names}."
                )

        results = []
        if len(names) == 0:
            return results
        if len(names) != len(items):
            raise RuntimeError('Can not match "names" and "items" lists, len(names) != len(items)')

        for batch in batched(list(zip(names, items, metas))):
            images = []
            for name, item, meta in batch:
                item_tuple = func_item_to_kv(item)
                images.append(
                    {
                        ApiField.NAME: name,
                        item_tuple[0]: item_tuple[1],
                        ApiField.META: meta,
                    }
                )
            response = self._api.post(
                "point-clouds.bulk.add",
                {ApiField.DATASET_ID: dataset_id, ApiField.POINTCLOUDS: images},
            )
            if progress_cb is not None:
                progress_cb(len(images))

            results.extend([self._convert_json_info(item) for item in response.json()])
        name_to_res = {img_info.name: img_info for img_info in results}
        ordered_results = [name_to_res[name] for name in names]

        return ordered_results



from typing import Dict, List

from supervisely import is_community
from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget


class AgentSelector(Widget):
    class Routes:
        VALUE_CHANGED = "value_changed"

    def __init__(
        self,
        team_id: int,
        show_only_gpu: bool = False,
        show_only_running: bool = True,
        compact: bool = False,
        widget_id=None,
    ):
        self._team_id = team_id
        self._show_any_status = not show_only_running
        self._show_public = True
        self._show_only_gpu = show_only_gpu
        self._check_network_status = True
        self._compact = compact

        self._is_community = is_community()

        self._changes_handled = False
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict:
        return {
            "teamId": self._team_id,
            "isCommunity": self._is_community,
        }

    def get_json_state(self) -> Dict:
        return {
            "agentId": None,
            "options": {
                "small": self._compact,
                "anyStatus": self._show_any_status,
                "showPublic": self._show_public,
                "needGpu": self._show_only_gpu,
                "checkAgentNetwork": self._check_network_status,
            },
        }

    def get_value(self) -> int:
        return StateJson()[self.widget_id]["agentId"]

    def value_changed(self, func):
        route_path = self.get_route_path(AgentSelector.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        async def _click():
            res = self.get_value()
            func(res)

        return _click



import supervisely.nn.inference as inference
from supervisely.nn.prediction_dto import PredictionMask, PredictionBBox, Prediction, PredictionSegmentation, PredictionKeypoints, PredictionCuboid3d
from supervisely.nn.prediction_dto import (
    PredictionMask,
    PredictionBBox,
    Prediction,
    PredictionSegmentation,
    PredictionKeypoints,
)

import supervisely.nn.checkpoints as checkpoints
from supervisely.nn.checkpoints.checkpoint import CheckpointInfo



import os

from pathlib import Path
from typing import List, Optional

from supervisely import (
    Api,
    PointcloudEpisodeAnnotation,
    ProjectMeta,
    batched,
    generate_free_name,
    is_development,
    logger,
)
from supervisely.api.module_api import ApiField
from supervisely.convert.base_converter import BaseConverter
from supervisely.io.json import load_json_file
from supervisely.io.fs import get_file_ext, get_file_name_with_ext
from supervisely.pointcloud.pointcloud import ALLOWED_POINTCLOUD_EXTENSIONS


class PointcloudEpisodeConverter(BaseConverter):
    allowed_exts = ALLOWED_POINTCLOUD_EXTENSIONS

    class Item(BaseConverter.BaseItem):
        def __init__(
            self,
            item_path,
            frame_number: int,
            ann_data: Optional[str] = None,
            related_images: Optional[list] = None,
            custom_data: Optional[dict] = None,
        ):
            self._path = item_path
            self._frame_number = frame_number
            self._ann_data = ann_data
            self._related_images = related_images if related_images is not None else []
            self._type = "point_cloud_episode"
            self._custom_data = custom_data if custom_data is not None else {}

        @property
        def frame_number(self) -> int:
            return self._frame_number

        def create_empty_annotation(self) -> PointcloudEpisodeAnnotation:
            return PointcloudEpisodeAnnotation()

        def set_related_images(self, related_images: dict) -> None:
            self._related_images.append(related_images)

    def __init__(self, input_data: str, labeling_interface: str):
        self._input_data: str = input_data
        self._items: List[self.Item] = []
        self._meta: ProjectMeta = None
        self._annotation = None
        self._frame_pointcloud_map = None
        self._frame_count = None
        self._labeling_interface: str = labeling_interface
        self._converter = self._detect_format()

    @property
    def format(self):
        return self._converter.format

    @property
    def frame_count(self):
        if self._frame_count is None:
            self._frame_count = len(self._items)
        return self._frame_count

    @property
    def ann_ext(self):
        return None

    @property
    def key_file_ext(self):
        return None

    @staticmethod
    def validate_ann_file(ann_path, meta=None):
        return False

    def upload_dataset(
        self,
        api: Api,
        dataset_id: int,
        batch_size: int = 1,
        log_progress=True,
    ):
        """Upload converted data to Supervisely"""

        meta, renamed_classes, renamed_tags = self.merge_metas_with_conflicts(
            api, dataset_id
        )

        existing_names = set(
            [pcde.name for pcde in api.pointcloud_episode.get_list(dataset_id)]
        )

        if log_progress:
            progress, progress_cb = self.get_progress(
                self.items_count, "Uploading pointcloud episodes..."
            )
        else:
            progress_cb = None

        frame_to_pointcloud_ids = {}
        for batch in batched(self._items, batch_size=batch_size):
            item_names = []
            item_paths = []
            item_metas = []
            for item in batch:
                item.name = generate_free_name(
                    existing_names, item.name, with_ext=True, extend_used_names=True
                )
                item_names.append(item.name)
                item_paths.append(item.path)

            pcd_infos = api.pointcloud_episode.upload_paths(
                dataset_id,
                item_names,
                item_paths,
            )
            pcd_ids = [pcd_info.id for pcd_info in pcd_infos]

            for item, pcd_id in zip(batch, pcd_ids):
                frame_to_pointcloud_ids[item._frame_number] = pcd_id
                item_metas.append({"frame": item._frame_number})
                rimg_infos = []
                camera_names = []
                if len(item._related_images) > 0:
                    img_paths, rimg_ann_paths = list(zip(*item._related_images))
                    rimg_hashes = api.pointcloud_episode.upload_related_images(
                        img_paths
                    )
                    for img_ind, (img_hash, rimg_ann_path) in enumerate(
                        zip(rimg_hashes, rimg_ann_paths)
                    ):
                        meta_json = load_json_file(rimg_ann_path)
                        try:
                            if ApiField.META not in meta_json:
                                raise ValueError(
                                    "Related image meta not found in json file."
                                )
                            if ApiField.NAME not in meta_json:
                                raise ValueError(
                                    "Related image name not found in json file."
                                )
                            if "deviceId" not in meta_json[ApiField.META].keys():
                                camera_names.append(f"CAM_{str(img_ind).zfill(2)}")
                            else:
                                camera_names.append(
                                    meta_json[ApiField.META]["deviceId"]
                                )
                            rimg_infos.append(
                                {
                                    ApiField.ENTITY_ID: pcd_id,
                                    ApiField.NAME: meta_json[ApiField.NAME],
                                    ApiField.HASH: img_hash,
                                    ApiField.META: meta_json[ApiField.META],
                                }
                            )
                            api.pointcloud.add_related_images(rimg_infos, camera_names)
                        except Exception as e:
                            logger.warn(
                                f"Failed to upload related image or add it to pointcloud episo: {repr(e)}"
                            )
                            continue

            if log_progress:
                progress_cb(len(batch))

        if self.items_count > 0:
            ann = self.to_supervisely(
                self._items[0], meta, renamed_classes, renamed_tags
            )
            api.pointcloud_episode.annotation.append(
                dataset_id, ann, frame_to_pointcloud_ids
            )

        if log_progress:
            if is_development():
                progress.close()
        logger.info(f"Dataset ID:{dataset_id} has been successfully uploaded.")
